#ifndef BLYNKCLOUDCLIENT_H
#define BLYNKCLOUDCLIENT_H

#include <QObject>
#include <QNetworkAccessManager>

class CBlynkCloudClient : public QObject
{
    Q_OBJECT

public:
    CBlynkCloudClient();

    void TestBlynk();
    void UpdateChaletVoltage(float Voltage);
    void UpdateChaletCurrent(int Current);
    void UpdateChaletWifiStatus(int Status);
    void UpdateChaletInverterStatus(int Status);
    void UpdateChaletStatusLed(int Status);

    int GetChaletWifiToggleSwitchStatus();

    int mDummy;
private:
    QNetworkAccessManager *mBlynkSocket;

public slots:
    void BlynkServerFinished(QNetworkReply*);
    void sslErrors(QNetworkReply*,QList<QSslError>);
};

#endif // BLYNKCLOUDCLIENT_H
